#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <io.h>
#include "memhandler.h"

#define DEBUG

struct Tag
{
	char* name;
	char* pattern;
};
Tag* tags;
int allocated_tags=0;
int num_user_tags=0;
const int user_tag_base=64; // number of system tags allowed
char* curfile;

enum
{
	SysTag_include=1,
	SysTag_comment,
	SysTag_section,
	SysTag_end,
	SysTag_put,
	SysTag_text,
	SysTag_define,
	SysTag_param,
	SysTag_endef,
	SysTag_hidden,
	SysTag_date
};

enum
{
	Mode_normal=0,
	Mode_define,
	Mode_section,
	Mode_text
};

int modestack[32];
int curmode=0;

struct BHandle
{
	char* sptr;
	char* ptr;
	bool eof;
	int allocsize;
};

struct Section
{
	char* name;
	char* text;
};

BHandle section_buffer;
int num_sections=1;
Section sections[1024]; // shouldn't be fixed size... bleh
int defining_section=0; // index of section being defined
bool expand_sections;
int tags_encountered=0;

BHandle usertag_buffer;
char usertag_name[64];

char tagparams[1024];
int tagparam_pos[32];

bool parse_error=false;
bool any_errors=false;
bool file_hidden=false;

char readchar(BHandle* buffer)
{
	char ch=*buffer->ptr;
	buffer->ptr++;
	if(ch=='\0')
	{
		buffer->eof=true;
		ch=' '; // return value shouldn't really be read in this case, but spaces are at least pretty harmless
	}
	return ch;
}

void EnterMode(int newmode)
{
	modestack[curmode++]=newmode;
}

void LeaveMode()
{
	curmode--;
}

bool inMode(int mode)
{
	for(int i=0;i<curmode;i++)
		if(modestack[i]==mode)
			return true;
	return false;
}

bool topMode(int mode)
{
	if(curmode>0)
		if(modestack[curmode-1]==mode)
			return true;
	return false;
}

bool empty_text=false;

void writestring(BHandle* output, char* string);

void writechar(BHandle* buffer, char ch)
{
	if(ch=='[')
		tags_encountered++; // we might need a new pass

	if(inMode(Mode_section))
		buffer=&section_buffer;
	if(topMode(Mode_define))
		buffer=&usertag_buffer;
	if(topMode(Mode_text) && ch=='\n' && !empty_text)
		writestring(buffer, "<br>");
	if(topMode(Mode_text) && ch!=' ')
		empty_text=false; // will occur at the first linebreak

	int size=buffer->ptr-buffer->sptr;
	if(size>=buffer->allocsize)
	{
		buffer->allocsize*=2;
		buffer->sptr=(char*)realloc(buffer->sptr, buffer->allocsize);
		buffer->ptr=buffer->sptr+size;
	}
	*buffer->ptr=ch;
	buffer->ptr++;
}

void writestring(BHandle* output, char* string)
{
	int length=strlen(string);
	for(int i=0;i<length;i++)
		writechar(output, string[i]);
}

void InitOutput(BHandle* buffer, int size)
{
	buffer->allocsize=size;
	buffer->sptr=(char*)realloc(NULL, buffer->allocsize);
	buffer->ptr=buffer->sptr;
}

char* LoadFile(char* filename)
{
	FILE* file=fopen(filename, "r");
	if(file==NULL)
		return NULL;
	int allocsize=1024;
	char* mem=(char*)realloc(NULL, allocsize); // caller responsible for freeing
	int pos=0;
	while(!feof(file))
	{
		if(pos>=allocsize)
		{
			allocsize*=2;
			mem=(char*)realloc(mem, allocsize);
		}
		fread(&mem[pos++], 1, 1, file);
	}
	mem[pos-1]='\0'; // eof marker
	fclose(file);
	return mem;
}

void SaveFile(BHandle* buffer, char* filename)
{
	FILE* file=fopen(filename, "w");
	buffer->ptr=buffer->sptr;
	buffer->eof=false;
	char ch=readchar(buffer);
	while(!buffer->eof)
	{
		fwrite(&ch, 1, 1, file);
		ch=readchar(buffer);
	}
	fclose(file);
}

bool CompareFile(BHandle* buffer, char* filename)
{
	FILE* file=fopen(filename, "r");
	if(file==NULL)
		return false;
	buffer->ptr=buffer->sptr;
	buffer->eof=false;
	int fcount=0;
	int bcount=0;
	char filech=' ';
	char buffch=readchar(buffer);
	fread(&filech, 1, 1, file);
	while(!(buffer->eof && feof(file)))
	{
		if(buffch=='') // special character to stop comparison, ugly fix for avoiding the date footer to always trigger file refresh
		{
			fclose(file);
			return true;
		}
		if(!buffer->eof && !feof(file) && filech!=buffch)
		{
			fclose(file);
			return false;
		}
		if(!buffer->eof)
		{
			bcount++;
			buffch=readchar(buffer);
		}
		else
			buffch='\0';
		if(!feof(file))
		{
			fcount++;
			fread(&filech, 1, 1, file);
		}
		else
			filech='\0';
	}
	if(buffer->eof && feof(file))
	{
		fclose(file);
		return true;
	}
	fclose(file);
	return false;
}

bool PrintError(char* message, ...)
{
	char string[256];
	va_list args;
	va_start(args, message);
	vsprintf(string, message, args);
	va_end(args);
	printf(" ### error: %s\n", string);
	parse_error=true;
	any_errors=true;
	return false;
}

void DisplayTags()
{
	printf("tags:\n");
	for(int i=1;i<user_tag_base+num_user_tags;i++)
		if(tags[i].name!=NULL)
		{
			if(i<user_tag_base)
				printf("%i: [%s] (system tag)\n", i, tags[i].name);
			else
				printf("%i: [%s] - \"%s\"\n", i, tags[i].name, tags[i].pattern);
		}
}

void DisplaySections()
{
	printf("sections:\n");
	for(int i=1;i<num_sections;i++)
		printf("%i: \"%s\" (%i chars)\n", i, sections[i].name, strlen(sections[i].text));
}

void FreeSections()
{
	for(int i=1;i<num_sections;i++)
	{
		free(sections[i].text);
		free(sections[i].name);
	}
	num_sections=1;
}

int FindSection(char* name)
{
	for(int i=1;i<num_sections;i++)
		if(strcmp(sections[i].name, name)==0)
			return i;
	return 0;
}

int FindTag(char* name)
{
	for(int i=1;i<64+num_user_tags;i++)
		if(tags[i].name!=NULL && strcmp(tags[i].name, name)==0)
			return i;
	return 0;
}

bool CreateTag(int id, char* name, char* pattern)
{
	if(name==NULL)
		return PrintError("define - no tag name provided");
	bool addednew=true;
	int existing=FindTag(name);
	if(existing)
	{
		id=existing;
		mh_free(tags[id].name);
		mh_free(tags[id].pattern);
		addednew=false;
	}

	if(id>=allocated_tags)
	{
		allocated_tags*=2;
		tags=(Tag*)realloc(tags, allocated_tags*sizeof(Tag));
	}

	tags[id].name=(char*)mh_malloc((int)strlen(name)+1);
	strcpy(tags[id].name, name);

	if(pattern==NULL)
		tags[id].pattern=NULL;
	else
	{
		int plength=(int)strlen(pattern);
		tags[id].pattern=(char*)mh_malloc(plength+1);
		strcpy(tags[id].pattern, pattern);
		for(int i=plength-1;i>=0;i--) // trim trailing spaces in pattern string
			if(tags[id].pattern[i]!=' ')
			{
				tags[id].pattern[i+1]='\0';
				break;
			}
	}

	return addednew;
}

int FindMonth(char *string)
{
	char months[12][4]={"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
	for(int i=0;i<12;i++)
	{
		for(int j=0;j<3;j++)
			if(string[j]!=months[i][j])
				break;
		if(j==3)
			return i+1;
	}
	return -1;
}

void WriteDate(BHandle* output)
{
	char *timestr;
	time_t curtime;
	char months[13][10]={ "January",
						"February",
						"March",
						"April",
						"May",
						"June",
						"July",
						"August",
						"September",
						"October",
						"November",
						"December",
						"Unknober"};

	// Find current date
	time(&curtime);
	timestr=ctime(&curtime);
	
	char monthshort[4];
	monthshort[0]=timestr[4];
	monthshort[1]=timestr[5];
	monthshort[2]=timestr[6];
	monthshort[3]='\0';
	int mi;
	for(mi=0;mi<12;mi++)
	{
		int i;
		for(i=0;i<3;i++)
			if(monthshort[i]!=months[mi][i])
				break;
		if(i==3) // all matched
			break;
	}
	char daystr[4];
	timestr[10]='\0';
	sscanf(&timestr[8], "%s", daystr);
	char yearstr[5];
	sscanf(&timestr[19], "%s", yearstr);
	char datestring[32];
	sprintf(datestring, "%s %s &nbsp %s", months[mi], daystr, yearstr);
	writestring(output, datestring);
}

bool Parse(BHandle* input, int endtag, BHandle* output);

bool ExecuteTag(int id, char* params, int* param_pos, BHandle* output)
{
	if(topMode(Mode_define) && id!=SysTag_endef) // pass through all tags except [endef] in tag definitions
	{
		char string[256];
		if(params[0]=='\0')
			sprintf(string, "[%s]", tags[id].name);
		else
			sprintf(string, "[%s %s]", tags[id].name, params);
		writestring(output, string);
		return true;
	}
	if(id>=user_tag_base)
	{
		strcpy(tagparams, params);
		for(int i=0;i<32;i++)
			tagparam_pos[i]=param_pos[i];
		BHandle buffer;
		buffer.sptr=tags[id].pattern;
		buffer.ptr=buffer.sptr;
		buffer.eof=false;
		Parse(&buffer, 0, output);
		return true;
	}
	switch(id)
	{
	case SysTag_include:
	{
		char filename[256];
		sprintf(filename, "pages\\%s.txt", params);
		BHandle buffer;
		char* ptr_store=LoadFile(filename);
		buffer.ptr=ptr_store;
		buffer.eof=false;
		bool wasok=Parse(&buffer, 0, output); // parse until eof
		free(ptr_store);
		return wasok;
	}
	case SysTag_define:
	{
		strcpy(usertag_name, params);
		InitOutput(&usertag_buffer, 256);
		EnterMode(Mode_define);
		return true;
	}
	case SysTag_endef:
		// don't do anything in particular, other parts of the code will use this to detect end of tag definition
		if(topMode(Mode_define))
		{
			writechar(output, '\0');
			bool addednew=CreateTag(user_tag_base+num_user_tags, usertag_name, usertag_buffer.sptr);
			free(usertag_buffer.sptr);
			if(addednew)
				num_user_tags++;
		}
		else
			PrintError("[endef] without prior [define]");
		LeaveMode();
		return true;
	case SysTag_param:
	{
		// this should turn up while reading tag definition... hmm
		// actually this should probably be ignored here and handled explicitly in a function for macro expansion (disallowing recursive macros...)
		if(topMode(Mode_define)) // pass through tag unchanged
		{
			char string[256];
			if(params[0]=='\0')
				sprintf(string, "[%s]", tags[id].name);
			else
				sprintf(string, "[%s %s]", tags[id].name, params);
			writestring(output, string);
		}
		else // insert relevant parameter
		{
			int pid=-1;
			if(params[0]>='1' && params[0]<='9')
				pid=params[0]-'1';
			if(pid==-1)
				writestring(output, tagparams);
			else
			{
				if(pid==0) // only allow two parameters at most, second one spans the rest of the string
				{
					char* ptr=&tagparams[tagparam_pos[pid]];
					while(*ptr!=' ' && *ptr!='\0')
						writechar(output, *ptr++);
				}
				else
					writestring(output, &tagparams[tagparam_pos[pid]]); // only allow two parameters at most, second one spans the rest of the string
			}
		}
		return true;
	}
	case SysTag_hidden:
		if(curmode==1) // if encountered in top file, make note about it
			file_hidden=true;
		return true;
	case SysTag_section:
	{
		// start storing section text
		if(defining_section!=0)
		{
			PrintError("section \"%s\" defined inside another section", params);
			return false;
		}
		defining_section=num_sections;
		int si=FindSection(params);
		if(si)
		{
			// existing section, redefine it
			free(sections[si].name);
			free(sections[si].text);
			defining_section=si;
		}
		InitOutput(&section_buffer, 1024);
		sections[defining_section].name=(char*)malloc(strlen(params)+1);
		strcpy(sections[defining_section].name, params);
		EnterMode(Mode_section);
		return true;
	}
	case SysTag_end:
		// end of whatever is open, could check parameters for safety
		if(topMode(Mode_section))
		{
			// store section text
			writechar(output, '\0');
			sections[defining_section].text=section_buffer.sptr;
			if(defining_section==num_sections)
				num_sections++;
			defining_section=0;
		}
		LeaveMode();
		return true;
	case SysTag_put:
		if(!expand_sections)
		{
			// in first pass the whole tag should be printed out as-is
			char string[256];
			sprintf(string, "[%s %s]", tags[id].name, params);
			writestring(output, string);
		}
		else
		{
			int si=FindSection(params);
			if(!si)
			{
				PrintError("undefined section \"%s\"", params);
				writestring(output, params);
			}
			else
			{
				writestring(output, sections[si].text);

				tags_encountered--; // this tag was eliminated
			}
			// in second pass, expand section <params> and print resulting character stream to output buffer
			// ...
		}
		return true;
	case SysTag_date:
		WriteDate(output);
		return true;
	case SysTag_text:
		// start auto-htmlized text mode, insert <br> at line breaks etc
		EnterMode(Mode_text);
		empty_text=true; // don't start adding <br> until we hit actual content (to avoid always adding a <br> right after the tag)
		return true;
	case SysTag_comment:
		// do nothing
		return true;
	}
	parse_error=true;
	return PrintError("unimplemented tag [%s] (?)", tags[id].name);
}

char* paramtext=NULL;
int paramtext_allocsize=0;

bool ReadTag(BHandle* input, int endtag, BHandle* output)
{
	char tname[64];
	int tnpos=0;
	int ptpos=0;
	int read_parameter=-2;
	bool param_whitespace=false;
	int param_pos[32];
	for(int i=0;i<32;i++)
		param_pos[i]=0;
	char ch='\0';
	do
	{
		ch=readchar(input);

		if(ch!=']')
		{
			if(read_parameter==-1 && ch!=' ')
				read_parameter=0; // start reading parameters for real
			if(read_parameter>-1)
			{
				if(ch==' ')
					param_whitespace=true;
				else if(param_whitespace) // we just came in from whitespace
				{
					param_whitespace=false;
					if(read_parameter<32)
						param_pos[read_parameter++]=ptpos; // this is the first character of this next parameter
				}
				if(ptpos>=paramtext_allocsize)
				{
					paramtext_allocsize*=2;
					paramtext=(char*)realloc(paramtext, paramtext_allocsize);
				}
				if(ch!=']')
					paramtext[ptpos++]=ch;
			}
			if(read_parameter==-2)
			{
				if(ch==' ' || tnpos==63) // end of tag name
				{
					tname[tnpos]='\0';
					read_parameter=-1; // start reading parameter string after whitespace
					param_whitespace=true;
				}
				else
					tname[tnpos++]=ch; // accumulate tag name characters
			}
		}
	} while(ch!=']' && !input->eof);
	tname[tnpos]='\0';
	paramtext[ptpos]='\0';
	param_pos[read_parameter]=ptpos+1; // so that inserting a '\0' before it won't break the previous parameter
	// lookup tag
	int tagi=FindTag(tname);
	if(!tagi)
	{
		PrintError("undefined tag [%s]", tname);
		parse_error=true;
		return false;
	}
	// execute it
	ExecuteTag(tagi, paramtext, param_pos, output);

	if(tagi==endtag) // this tag is the end condition, tell caller (parse) that they're done
		return true;
	return false;
}

bool Parse(BHandle* input, int endtag, BHandle* output)
{
	bool emptyline=true;
	int root_mode=curmode;

	EnterMode(Mode_normal);

	while(!input->eof)
	{
		char ch=' ';
		ch=readchar(input);
		if(input->eof)
			break;
		bool done=false;
		if(ch=='[')
		{
			tags_encountered++; // we might need a new pass
			done=ReadTag(input, endtag, output);
		}
		else if(topMode(Mode_normal) || topMode(Mode_section) || topMode(Mode_define))
		{
			if(ch!=' ' && ch!='\n')
				emptyline=false;
			if(!emptyline)
				writechar(output, ch);
			if(ch=='\n') // starting new line
				emptyline=true;
		}
		else if(topMode(Mode_text))
			writechar(output, ch);

		if(done || file_hidden)
			break;
	}

	LeaveMode();

	if(curmode!=root_mode)
	{
		PrintError("missing [end]");
		return false;
	}

	return true;
}

bool Generate(char* filename)
{
	printf("parsing \"%s\"\n", filename);
	curfile="somefile";
	BHandle inbuffer;
	inbuffer.sptr=LoadFile(filename);
	if(inbuffer.sptr==NULL)
	{
		printf("file not found\n", filename);
		return false;
	}
	inbuffer.ptr=inbuffer.sptr;
	inbuffer.eof=false;
	BHandle outbuffer;
	InitOutput(&outbuffer, 1024);
	curmode=0;
	parse_error=false;
	file_hidden=false;
	expand_sections=false;
	Parse(&inbuffer, 0, &outbuffer);
	writechar(&outbuffer, '\0');
	free(inbuffer.sptr);

	if(parse_error)
		printf("errors in pass 1\n");
	else if(!file_hidden)
	{
		int curpass=1;
		do
		{
			curpass++;
			inbuffer.sptr=outbuffer.sptr;
			inbuffer.ptr=inbuffer.sptr;
			inbuffer.eof=false;
			InitOutput(&outbuffer, 1024);
			curmode=0;
			parse_error=false;
			expand_sections=true;
			tags_encountered=0;
			Parse(&inbuffer, 0, &outbuffer);
			writechar(&outbuffer, '\0');
			free(inbuffer.sptr);
		} while(tags_encountered>0 && !parse_error); // keep passing over the code until no more tags are left
		if(parse_error)
			printf("errors in pass %i\n", curpass);
		else
		{
			char outname[256];
			int i=0;
			for(i=strlen(filename)-1;i>=0;i--)
			{
				if(filename[i]=='.')
					filename[i]='\0';
				if(filename[i]=='\\')
				{
					i+=1;
					break;
				}
			}
			strcpy(outname, &filename[i]);
			strcat(outname, ".html");
			if(!CompareFile(&outbuffer, outname))
			{
				printf("saving \"%s\"\n", outname);
				SaveFile(&outbuffer, outname);
			}
		}
	}
	free(outbuffer.sptr);
	FreeSections();
	return !parse_error;
}

void GenerateAll()
{
	any_errors=false;

	// iterate through page directory
	printf("regenerating all files\n");
	_finddata_t finfo;
	char pattern[256];
	strcpy(pattern, "pages\\*.txt");
	intptr_t handle=_findfirst(pattern, &finfo);
	if(handle==-1)
	{
		printf("no pages found\n");
		_findclose(handle);
		return;
	}

	char filename[256];
	for(;;)
	{
		strcpy(filename, "pages\\");
		strcat(filename, finfo.name);
		if(!Generate(filename))
			break;
		if(_findnext(handle, &finfo)==-1)
			break;
	}
	_findclose(handle);
	if(any_errors)
		printf("there were errors\n");
	else
		printf("all files are up to date\n");
}

struct FileEntry
{
	char* name;
	unsigned int lastchanged; // unix time
};

FileEntry* fentries;
int fentries_num;
int fentries_numalloc;
bool batched_any;

void LoadRecord(char* filename)
{
	fentries_num=0;
	FILE* record=fopen(filename, "rb");
	if(record!=NULL)
	{
		fread(&fentries_num, 1, sizeof(int), record);
		fentries_numalloc=fentries_num+128;
		fentries=(FileEntry*)realloc(NULL, fentries_numalloc*sizeof(FileEntry));
		for(int i=0;i<fentries_num;i++)
		{
			fread(&fentries[i], 1, sizeof(FileEntry), record);
			int namelen=0;
			fread(&namelen, 1, sizeof(int), record);
			fentries[i].name=(char*)malloc(namelen+1);
			fread(fentries[i].name, namelen, 1, record);
			fentries[i].name[namelen]='\0';
		}
		fclose(record);
	}
	else
	{
		fentries_numalloc=128;
		fentries=(FileEntry*)realloc(NULL, fentries_numalloc*sizeof(FileEntry));
	}
}

void SaveRecord(char* filename)
{
	FILE* record=fopen(filename, "wb");
	if(record!=NULL)
	{
		fwrite(&fentries_num, 1, sizeof(int), record);
		for(int i=0;i<fentries_num;i++)
		{
			fwrite(&fentries[i], 1, sizeof(FileEntry), record);
			int namelen=strlen(fentries[i].name);
			fwrite(&namelen, 1, sizeof(int), record);
			fwrite(fentries[i].name, namelen, 1, record);
		}
		fclose(record);
	}
}

FileEntry* FindEntry(char* name)
{
	for(int i=0;i<fentries_num;i++)
		if(strcmp(fentries[i].name, name)==0)
			return &fentries[i];
	return NULL;
}

void AddEntry(char* name, unsigned int lastchanged)
{
	FileEntry* entry=FindEntry(name);
	if(entry==NULL)
	{
		if(fentries_num>=fentries_numalloc)
		{
			fentries_numalloc*=2;
			fentries=(FileEntry*)realloc(fentries, fentries_numalloc*sizeof(FileEntry));
		}
		entry=&fentries[fentries_num];
		entry->name=(char*)malloc(strlen(name)+1);
		fentries_num++;
	}
	strcpy(entry->name, name);
	entry->lastchanged=lastchanged;
}

void BatchChanged(char* path, char* wildcard, FILE* batch)
{
	_finddata_t finfo;
	char pattern[256];
	strcpy(pattern, path);
	strcat(pattern, wildcard);
	intptr_t handle=_findfirst(pattern, &finfo);
	if(handle==-1)
	{
		printf("no files found\n");
		_findclose(handle);
		return;
	}

	char filename[256];
	for(;;)
	{
		if(!(finfo.attrib&_A_SUBDIR))
		{
			strcpy(filename, path);
			strcat(filename, finfo.name);
			FileEntry* fentry=FindEntry(filename);
			if(fentry==NULL || fentry->lastchanged!=finfo.time_write)
			{
				printf("/%s\n", filename);
				if(path[0]!='\0')
					fprintf(batch, "put \"%s\" \"%s\"\n", filename, filename);
				else
					fprintf(batch, "put \"%s\"\n", filename);
				AddEntry(filename, finfo.time_write);
				batched_any=true;
			}
		}
		if(_findnext(handle, &finfo)==-1)
			break;
	}
	_findclose(handle);
}

char sftpstring[256]="";
char sftpdir[256]="";

void Upload()
{
	printf("searching for changed files\n");

	// load file record
	LoadRecord("record.dat");

	FILE* batch=fopen("batch.txt", "w");
	if(batch==NULL)
	{
		printf("couldn't open batch.txt for writing\n");
		return;
	}
	fprintf(batch, "cd %s\n", sftpdir);

	batched_any=false;
	BatchChanged("", "*.html", batch);
	BatchChanged("images/", "*.*", batch);
	BatchChanged("thumbnails/", "*.*", batch);
	BatchChanged("files/", "*.*", batch);
	
	fclose(batch);

	if(batched_any)
	{
		printf("connecting to sftp server...\n");
		char cmdstring[256];
		sprintf(cmdstring, "psftp %s -b batch.txt", sftpstring);
		system(cmdstring);
		printf("upload complete\n");

		// save updated file record
		SaveRecord("record.dat");
	}
	else
		printf("no files to upload\n");

	// free file record
	for(int i=0;i<fentries_num;i++)
		free(fentries[i].name);
	free(fentries);
	
	system("del batch.txt");
}

void InstallServer()
{
	FILE* batch=fopen("batch.txt", "w");
	if(batch==NULL)
	{
		printf("couldn't open batch.txt for writing\n");
		return;
	}
	fprintf(batch, "cd %s\n", sftpdir);
	fprintf(batch, "mkdir images\n");
	fprintf(batch, "mkdir thumbnails\n");
	fprintf(batch, "mkdir files\n");
	fclose(batch);

	printf("connecting to sftp server...\n");
	char cmdstring[256];
	sprintf(cmdstring, "psftp %s -b batch.txt", sftpstring);
	system(cmdstring);
	printf("server installation complete\n");
	
	system("del batch.txt");
}

void PrintHelp()
{
	printf("\n");
	printf("commands:\n");
	printf(" g, generate  - generate html pages and report any errors\n");
	printf(" u, upload    - upload modified files to server\n");
	printf(" i, install   - install spcms on server (create directories)\n");
	printf(" f, psftp     - manual access to server\n");
	printf(" h, help      - show available commands\n");
	printf(" x, exit      - exit program\n");
	printf("\n");
}

int main(int argc, char** argv)
{
	int auto_commands=0;
	const int auto_generate=0x01;
	const int auto_upload=0x02;
	const int auto_install=0x03;

	if(argc==1)
	{
		printf("\n\n\nwelcome to spcms\n");
		PrintHelp();
	}
	else
	{
		for(int i=1;i<argc;i++)
		{
			if(strcmp(argv[i], "-g")==0)
				auto_commands|=auto_generate;
			if(strcmp(argv[i], "-u")==0)
				auto_commands|=auto_upload;
			if(strcmp(argv[i], "-i")==0)
				auto_commands|=auto_install;
		}
	}

	paramtext_allocsize=128;
	paramtext=(char*)realloc(NULL, paramtext_allocsize);

	allocated_tags=128;
	tags=(Tag*)realloc(NULL, allocated_tags*sizeof(Tag));
	for(int i=0;i<user_tag_base;i++)
	{
		tags[i].name=NULL;
		tags[i].pattern=NULL;
	}

	CreateTag(SysTag_include, "include", NULL);
	CreateTag(SysTag_comment, "//", NULL);
	CreateTag(SysTag_section, "section", NULL);
	CreateTag(SysTag_end, "end", NULL);
	CreateTag(SysTag_put, "put", NULL);
	CreateTag(SysTag_text, "text", NULL);
	CreateTag(SysTag_define, "define", NULL);
	CreateTag(SysTag_param, "param", NULL);
	CreateTag(SysTag_endef, "endef", NULL);
	CreateTag(SysTag_hidden, "hidden", NULL);
	CreateTag(SysTag_date, "date", NULL);

	FILE* cfg=fopen("config.txt", "r");
	if(cfg!=NULL)
	{
		fscanf(cfg, "sftp: %s\n", sftpstring);
		fscanf(cfg, "dir: %s\n", sftpdir);
		fclose(cfg);
	}

	if(auto_commands!=0) // execute commandline arguments and quit
	{
		if(auto_generate)
			GenerateAll();
		if(auto_install)
			InstallServer();
		if(auto_upload)
			Upload();
		free(tags);
		free(paramtext);
		mh_cleanup();
		return 0;
	}

	bool done=false;
	for(;;)
	{
		printf("> ");
		int ci=0;
		char user_command[256];
		do
		{
			user_command[ci++]=getchar();
		} while(user_command[ci-1]!='\n');
		user_command[ci-1]='\0';
		bool unknown=true;
		if(strcmp(user_command, "g")==0) { GenerateAll(); unknown=false; }
		if(strcmp(user_command, "u")==0 || strcmp(user_command, "f")==0 || strcmp(user_command, "i")==0) // some sftp action
		{
			bool cmd_upload=(strcmp(user_command, "u")==0);
			bool cmd_manual=(strcmp(user_command, "f")==0);
			bool cmd_install=(strcmp(user_command, "i")==0);
			bool cfgchanged=false;
			if(sftpstring[0]=='\0')
			{
				printf("enter sftp open string (host or user@host): ");
				char fc=getchar(); // read first char
				if(fc!='\n')
				{
					sftpstring[0]=fc;
					scanf("%s", &sftpstring[1]);
					if(strlen(sftpstring)>0)
						cfgchanged=true;
					getchar(); // read linebreak
				}
			}
			if(sftpdir[0]=='\0' && (cmd_upload || cmd_install))
			{
				printf("enter server directory (where index.html will go): ");
				char fc=getchar(); // read first char
				if(fc!='\n')
				{
					sftpdir[0]=fc;
					scanf("%s", &sftpdir[1]);
					if(strlen(sftpdir)>0)
						cfgchanged=true;
					getchar(); // read linebreak
				}
			}
			if(sftpstring[0]!='\0')
			{
				if(cmd_upload || cmd_install)
				{
					if(sftpdir[0]=='\0')
						printf("missing sftp server directory, check config.txt\n");
					else
					{
						if(cmd_upload)
							Upload();
						else
							InstallServer();
					}
				}
				else if(cmd_manual)
				{
					char cmdstring[256];
					sprintf(cmdstring, "psftp %s", sftpstring);
					system(cmdstring);
				}
				unknown=false;
			}
			else
				printf("missing sftp open string, check config.txt\n");
			if(cfgchanged)
			{
				FILE* cfg=fopen("config.txt", "w");
				if(cfg!=NULL)
				{
					fprintf(cfg, "sftp: %s\n", sftpstring);
					fprintf(cfg, "dir: %s\n", sftpdir);
					fclose(cfg);
				}
			}
		}
		if(strcmp(user_command, "h")==0) { PrintHelp(); unknown=false; }
		if(strcmp(user_command, "x")==0) { break; unknown=false; }
		if(unknown) printf("unknown command\n");
	}

	free(tags);
	free(paramtext);

	mh_cleanup();
	
	printf("have a good one\n");

	return 0;
}

